
' -----------------------------------------------------------------------------
'
'       TileMap.bmx
'
'       (C) 2006 Mike Wiering, Wiering Software
'       http://www.wieringsoftware.nl/
'
'       To be used with Tile Studio (http://tilestudio.sourceforge.net/)
'       Export with BMXPNG.tsd
'
' -----------------------------------------------------------------------------

SuperStrict


' -----------------------------------------------------------------------------
' replacement for "Mod" without negative numbers
' -----------------------------------------------------------------------------

Function SafeMod: Int (X: Int, Y: Int)
  Local Z: Int
  Z = X Mod Y
  If Z < 0 Then Z :+ Y
  Return Z
End Function


' -----------------------------------------------------------------------------
' TTileSet class
' -----------------------------------------------------------------------------

Type TTileSet

  Field Identifier: String
  Field TileSetImage: TImage
  Field TileW: Int
  Field TileH: Int
  Field Overlap: Int
  Field TileCount: Int
  Field Sequences: Int[][]
  Field TotalFrames: Int[]


  Method Init (ID: String, W: Int, H: Int, OL: Int, Count: Int, Seq: Int[][])
    Identifier = ID
    TileW = W
    TileH = H
    Overlap = OL
    TileCount = Count
    TileSetImage = LoadAnimImage ("incbin::" + ID + "TileSet.png", W, H, 0, Count)

    For Local k: Int = 0 Until Count
      Local pixmap: TPixmap = LockImage (TileSetImage, k)
      For Local j: Int = 0 Until ImageHeight (TileSetImage)
        For Local i: Int = 0 Until ImageWidth (TileSetImage)
          Local rgba: Int = ReadPixel (pixmap, i, j)
         ' rgba = rgba | $1F1F1F
         ' rgba = rgba & $FFC0C0C0 + $202020
          WritePixel (pixmap, i, j, rgba)
        Next
      Next
      UnlockImage (TileSetImage, k)
    Next

    Sequences = Seq
    TotalFrames = New Int[Len (Seq)]
    For Local n: Int = 0 To Len (Seq) - 1
      TotalFrames[n] = 0
      For Local i: Int = 0 To Len (Sequences[n]) / 3 - 1
        TotalFrames[n] :+ Sequences[n][3 * i + 1]
      Next
    Next
  End Method


  Method GetSequenceFrame: Int (n: Int, Time: Int)
    Local tile: Int = 0
    n = Abs (n) - 1
    If n < Len (Sequences) Then
      Local i: Int = Time Mod TotalFrames[n]
      While i >= Sequences[n][tile + 1]
        i :- Sequences[n][tile + 1]
        tile :+ 3
      Wend
      tile = Sequences[n][tile]
    End If
    Return tile
  End Method


End Type



' -----------------------------------------------------------------------------
' TLayer class
' -----------------------------------------------------------------------------

Const MRP_NONE%     =  0   ' Map Repeat Patterns
Const MRP_TILE%     =  1
Const MRP_EXTEND%   =  2


Type TLayer

  Field TileSet: TTileSet
  Field ScrollX: Int
  Field ScrollY: Int
  Field MapW: Int
  Field MapH: Int
  Field MapData: Int[][]
  Field BoundMap: Int[][]
  Field ShiftX: Int
  Field ShiftY: Int
  Field TimeCounter: Int
  Field HorzRepPat: Int
  Field VertRepPat: Int
  Field Time: Float
  Field DT: Float
  Field TimeX: Float
  Field TimeY: Float


  Method Init (TS: TTileSet, Data: Int[][], W: Int, H: Int)
    TileSet = TS
    MapW = W
    MapH = H
    MapData = MapData[..H]
    For Local i: Int = 0 Until H
      MapData[i] = Data[i][..W]
    Next
    ShiftX = 0
    ShiftY = 0
    ScrollX = 0
    ScrollY = 0
    HorzRepPat = MRP_NONE
    VertRepPat = MRP_NONE
    Time = 0.0
    DT = 1.0
    TimeX = 0.0
    TimeY = 0.0
  End Method


  Method SetBoundMap (BndMap: Int[][])
    BoundMap = BndMap[..MapH]
    For Local i: Int = 0 Until MapH
      BoundMap[i] = BndMap[i][..MapW]
    Next
  End Method


  Method ScrollTo (X: Int, Y: Int)
    ScrollX = X
    ScrollY = Y
  End Method


  Method GetMapSize (W: Int Var, H: Int Var)
    Local XSc: Float, YSc: Float
    GetScale (XSc, YSc)
    W = MapW * TileSet.TileW * XSc
    H = MapH * (TileSet.TileH - TileSet.Overlap) * YSc
  End Method


  Method FindPositionInMap: Int (X: Int Var, Y: Int Var, flags: Int Var)
    Local OldX: Int = X
    Local OldY: Int = Y
    Local i: Int
    Select HorzRepPat 
      Case MRP_NONE    If (X < 0) Or (X > MapW - 1) Then Return False
      Case MRP_TILE    X = SafeMod (X, MapW)
      Case MRP_EXTEND  If X < 0 Then X = 0 Else If X > MapW - 1 Then X = MapW - 1
    End Select
    Select VertRepPat 
      Case MRP_NONE    If (Y < 0) Or (Y > MapH - 1) Then Return False
      Case MRP_TILE    Y = SafeMod (Y, MapH)
      Case MRP_EXTEND  If Y < 0 Then Y = 0 Else If Y > MapH - 1 Then Y = MapH - 1
    End Select
    Return True
  End Method


  Method GetTile: Int (X: Int, Y: Int)
    Local tile: Int = 0
    Local flags: Int = 0
    If FindPositionInMap (X, Y, flags) Then
      tile = MapData[Y][X]
      If tile < 0 Then tile = TileSet.GetSequenceFrame (tile, TimeCounter)
      If tile = 0 Then flags = 0
    End If
    Return tile | flags
  End Method


  Method Draw ()
    Local ScaleX: Float
    Local ScaleY: Float
    GetScale (ScaleX, ScaleY)
    Local ScaledTileW: Int = Floor (ScaleX * TileSet.TileW)
    Local ScaledTileH: Int = Floor (ScaleY * TileSet.TileH)
    Local ScaledTileHO: Int = Floor (ScaleY * (TileSet.TileH - TileSet.Overlap))
    Local vpx: Int, vpy: Int, vpw: Int, vph: Int
    GetViewport (vpx, vpy, vpw, vph)
    Local xp: Int = vpx + ScrollX + ShiftX - Time * TimeX
    Local yp: Int = vpy + ScrollY + ShiftY - Time * TimeY
    Local scry: Int = -(ScaledTileHO - 1) 
    Local mapy: Int = Floor ((scry + yp) / ScaledTileHO)
    Local y: Int = mapy * ScaledTileHO - yp
    SetImageHandle TileSet.TileSetImage, TileSet.TileW / 2, TileSet.TileH / 2
    Local dx: Float = ScaleX * (TileSet.TileW) / 2
    Local dy: Float = ScaleY * (TileSet.TileH) / 2
    While y < vph
      Local scrx: Int = -(ScaledTileW - 1)  ' -1 ???
      Local mapx: Int = Floor ((scrx + xp) / ScaledTileW)
      Local x: Int = mapx * ScaledTileW - xp
      While x < vpw
        Local tile: Int = GetTile (mapx, mapy)
        If tile > 0 Then
          DrawImage (TileSet.TileSetImage, x + dx, y + dy, tile - 1)
        End If
        x :+ ScaledTileW
        mapx :+ 1
      Wend
      y :+ ScaledTileHO
      mapy :+ 1
    Wend
    TimeCounter :+ 1
    Time :+ DT
  End Method


End Type


